@@ -28,6 +28,8 @@
 #include "debug.h"
 #include "zscii.h"
 
+#include <signal.h>
+
 #define yylval debug_eval_lval
 #include "eval.h"
 
@@ -35,8 +37,15 @@ debug_breakpoint* debug_bplist       = NULL;
 int               debug_nbps         = 0;
 debug_routine*    debug_expr_routine = NULL;
 
-static int*       expr      = NULL;
-static int        expr_pos  = 0;
+int*			  debug_expr      = NULL;
+int				  debug_expr_pos  = 0;
+
+static debug_breakpoint_handler bp_handler = NULL;
+
+/*
+static debug_breakpoint* return_breakpoint;
+static ZFrame*			 return_frame;
+ */
 
 typedef struct debug_display
 {
@@ -49,6 +58,8 @@ typedef struct debug_display
 static int            ndisps = 0;
 static debug_display* dbdisp = NULL;
 
+static debug_address addr;
+
 /***                           ----// 888 \\----                           ***/
 
 /* The debugger console */
@@ -58,25 +69,18 @@ static int stepinto = 0;
 /* Action when a breakpoint occurs */
 void debug_run_breakpoint(ZDWord pc)
 {
-  debug_address addr;
   debug_breakpoint* bp;
   static int banner = 0;
   int x;
+  ZFrame* frame;
 
   bp = debug_get_breakpoint(pc);
-  if (bp == NULL)
-    {
-      display_sanitise();
-      display_printf("=! Unexpected breakpoint\n");
-      display_desanitise();
-      return;
-    }
 
-  if (bp->usage == 1 && bp->funcbp && !stepinto)
+  if (bp && bp->usage == 1 && bp->funcbp && !stepinto)
     return;
 
   addr = debug_find_address(pc);
-  if (bp->usage == 1 && bp->funcbp && 
+  if (bp && bp->usage == 1 && bp->funcbp && 
       (!stepinto || addr.routine->defn_fl == 0 || addr.routine->defn_fl == 255))
     {
       return;
@@ -91,12 +95,25 @@ void debug_run_breakpoint(ZDWord pc)
 	  x--;
 	}
     }
-
+  
+  /* Clear any return breakpoints */
+  if (machine.stack.current_frame) {
+	  for (frame = machine.stack.current_frame; frame != NULL; frame = frame->last_frame) {
+		  frame->break_on_return = 0;
+	  }
+  }
+  
+  stepinto = 0;
+  
+  if (bp_handler != NULL) {
+	  /* Run the handler instead of the standard debugging routines */
+	  (*bp_handler)(pc);
+	  return;
+  }
+  
   display_sanitise();
   display_printf("=\n");
 
-  stepinto = 0;
-
   /* Print a quick banner if we're just starting up... */
   if (banner == 0)
     {
@@ -134,8 +151,8 @@ void debug_run_breakpoint(ZDWord pc)
   /* Evaluate any display expressions */
   for (x=0; x<ndisps; x++)
     {
-      expr = dbdisp[x].expr;
-      expr_pos = 0;
+      debug_expr = dbdisp[x].expr;
+      debug_expr_pos = 0;
       debug_expr_routine = addr.routine;
       debug_error = NULL;
       debug_eval_parse();
@@ -172,7 +189,7 @@ void debug_run_breakpoint(ZDWord pc)
 	free(debug_eval_type);
       debug_eval_type = NULL;
     }
-
+  
   /* Process commands */
   while (1)
     {
@@ -248,8 +265,8 @@ void debug_run_breakpoint(ZDWord pc)
 
 	case 'd':
 	  debug_expr_routine = addr.routine;
-	  expr = cline + 1;
-	  expr_pos = 0;
+	  debug_expr = cline + 1;
+	  debug_expr_pos = 0;
 	  debug_error = NULL;
 	  debug_eval_parse();
 	  if (debug_error == NULL)
@@ -257,19 +274,19 @@ void debug_run_breakpoint(ZDWord pc)
 	      int len;
 	      char* disp;
 
-	      for (len=0; expr[len] != 0; len++);
+	      for (len=0; debug_expr[len] != 0; len++);
 	      
 	      disp = malloc(sizeof(char)*(len+1));
-	      for (len=0; expr[len] != 0; len++)
-		disp[len] = expr[len];
+	      for (len=0; debug_expr[len] != 0; len++)
+		disp[len] = debug_expr[len];
 	      disp[len] = 0;
 
 	      dbdisp = realloc(dbdisp, sizeof(debug_display)*(ndisps+1));
 	      dbdisp[ndisps].desc = disp;
 	      dbdisp[ndisps].expr = malloc(sizeof(int)*(len+1));
 	      
-	      for (len=0; expr[len] != 0; len++)
-		dbdisp[ndisps].expr[len] = expr[len];
+	      for (len=0; debug_expr[len] != 0; len++)
+		dbdisp[ndisps].expr[len] = debug_expr[len];
 	      dbdisp[ndisps].expr[len] = 0;
 	      dbdisp[ndisps].lastvalue = debug_eval_result;
 	      dbdisp[ndisps].erm = 0;
@@ -293,8 +310,8 @@ void debug_run_breakpoint(ZDWord pc)
 
 	case 'p':
 	  debug_expr_routine = addr.routine;
-	  expr = cline + 1;
-	  expr_pos = 0;
+	  debug_expr = cline + 1;
+	  debug_expr_pos = 0;
 	  debug_error = NULL;
 	  debug_eval_parse();
 	  if (debug_error == NULL)
@@ -443,6 +460,7 @@ int debug_set_breakpoint(int address,
 			 int funcbp)
 {
   debug_breakpoint* bp;
+  int pos;
 
   bp = debug_get_breakpoint(address);
   if (bp != NULL)
@@ -458,15 +476,32 @@ int debug_set_breakpoint(int address,
 #ifdef DEBUG
   printf_debug("Setting BP @ %04x\n", address);
 #endif
+  
+  pos = 0;
+  
+  /* Find the breakpoint we should insert this new one before */
+  while (pos < debug_nbps && debug_bplist[pos].address < address) {
+    pos++;
+  }
 
+  /* Add a new breakpoint */
   debug_bplist = realloc(debug_bplist,
 			 sizeof(debug_breakpoint)*(debug_nbps+1));
-  debug_bplist[debug_nbps].address   = address;
-  debug_bplist[debug_nbps].original  = machine.memory[address];
-  debug_bplist[debug_nbps].usage     = 1;
-  debug_bplist[debug_nbps].temporary = temporary;
-  debug_bplist[debug_nbps].funcbp    = funcbp;
+  
+  if (pos < debug_nbps) {
+    /* Move the breakpoints up */
+    memmove(debug_bplist + pos + 1, debug_bplist + pos,
+	    (debug_nbps-pos)*sizeof(debug_breakpoint));
+  }
 
+  /* Store the actual breakpoint */
+  debug_bplist[pos].address   = address;
+  debug_bplist[pos].original  = machine.memory[address];
+  debug_bplist[pos].usage     = 1;
+  debug_bplist[pos].temporary = temporary;
+  debug_bplist[pos].funcbp    = funcbp;
+
+  /* Add a breakpoint instruction (we use status_nop, as it's just one byte) */
   machine.memory[address] = 0xbc; /* status_nop, our breakpoint */
 
   debug_nbps++;
@@ -476,6 +511,28 @@ int debug_set_breakpoint(int address,
 
 debug_breakpoint* debug_get_breakpoint(int address)
 {
+  int top, middle, bottom;
+  
+  /* Binary search for the breakpoint (they are stored sorted by address) */
+  top = debug_nbps-1;
+  bottom = 0;
+  
+  while (top >= bottom) {
+    middle = (top + bottom) >> 1;
+    
+    if (debug_bplist[middle].address > address) {
+      /* Need to search the lower half */
+      top = middle-1;
+    } else if (debug_bplist[middle].address < address) {
+      /* Need to search the upper half */
+      bottom = middle + 1;
+    } else {
+      /* Just right */
+      return debug_bplist + middle;
+    }
+  }
+  
+  /*
   int x;
 
   for (x=0; x<debug_nbps; x++)
@@ -483,6 +540,7 @@ debug_breakpoint* debug_get_breakpoint(int address)
       if (debug_bplist[x].address == address)
 	return debug_bplist + x;
     }
+   */
 
   return NULL;
 }
@@ -536,20 +594,29 @@ static void debug_add_symbol(char* name,
     }
   storename[x] = 0;
 
-  if (hash_get(debug_syms.symbol, storename, strlen(name)) != NULL)
+  if (hash_get(debug_syms.symbol, (unsigned char*)storename, strlen(name)) != NULL)
     {
       display_printf("=? Symbol space clash - %s\n", name);
     }
   else
     debug_syms.nsymbols++;
   hash_store_happy(debug_syms.symbol,
-		   storename,
+		   (unsigned char*)storename,
 		   strlen(name),
 		   sym);
+  
+  sym->next = debug_syms.first_symbol;
+  debug_syms.first_symbol = sym;
 
   free(storename);
 }
 
+#ifdef REMOTE_BREAKPOINT
+static void debug_sigusr1(int sig) {
+	machine.force_breakpoint = 1;
+}
+#endif
+
 void debug_load_symbols(char* filename,
 			char* pathname)
 {
@@ -561,10 +628,23 @@ void debug_load_symbols(char* filename,
   int done;
   
   int x;
-
+  
   debug_routine* this_routine = NULL;
   debug_symbol* sym;
-
+  
+#ifdef REMOTE_BREAKPOINT
+  /* SIGUSR1 indicates that we should break ASAP */
+  struct sigaction oldact;
+	  
+  sigaction(SIGUSR1, NULL, &oldact);
+  
+  oldact.sa_flags |= SA_RESTART;
+  oldact.sa_flags &= ~(SA_NODEFER|SA_SIGINFO);
+  oldact.sa_handler = debug_sigusr1;
+  
+  sigaction(SIGUSR1, &oldact, NULL);
+#endif
+				
   size = get_file_size(filename);
   file = open_file(filename);
 
@@ -589,6 +669,8 @@ void debug_load_symbols(char* filename,
       free(db_file);
       return;
     }
+  
+  debug_syms.largest_object = 0;
 
   pos = 6;
 
@@ -630,11 +712,11 @@ void debug_load_symbols(char* filename,
 	    fl = malloc(sizeof(debug_file));
 
 	    fl->number = db_file[pos+1];
-	    fl->name = malloc(sizeof(char)*(strlen(db_file + pos + 2) + 1));
-	    strcpy(fl->name, db_file + pos + 2);
+	    fl->name = malloc(sizeof(char)*(strlen((char*)(db_file + pos + 2)) + 1));
+	    strcpy(fl->name, (char*)(db_file + pos + 2));
 	    pos += 3 + strlen(fl->name);
-	    fl->realname = malloc(sizeof(char)*(strlen(db_file + pos) + 1));
-	    strcpy(fl->realname, db_file + pos);
+	    fl->realname = malloc(sizeof(char)*(strlen((char*)(db_file + pos)) + 1));
+	    strcpy(fl->realname, (char*)(db_file + pos));
 	    pos += strlen(fl->realname) + 1;
 
 	    fl->data   = NULL;
@@ -658,7 +740,7 @@ void debug_load_symbols(char* filename,
 		  {
 		    int x;
 
-		    fl->data = read_block(fl_load, 0, fl_len);
+		    fl->data = (char*)read_block(fl_load, 0, fl_len);
 		    close_file(fl_load);
 		    fl->data = realloc(fl->data, sizeof(char)*(fl_len+2));
 		    fl->data[fl_len] = 0;
@@ -713,7 +795,7 @@ void debug_load_symbols(char* filename,
 	    debug_syms.files[fl->number] = *fl;
 
 	    hash_store_happy(debug_syms.file,
-			     fl->name,
+			     (unsigned char*)fl->name,
 			     strlen(fl->name),
 			     fl);
 	  }
@@ -775,6 +857,8 @@ void debug_load_symbols(char* filename,
 	    sym->data.object = o;
 	    debug_add_symbol(o.name,
 			     sym);
+	    
+	    if (o.number > debug_syms.largest_object) debug_syms.largest_object = o.number;
 	  }
 	  break;
 	  
@@ -1054,6 +1138,8 @@ void debug_load_symbols(char* filename,
 		if (strcmp(name, "code area") == 0)
 		  {
 		    debug_syms.codearea = address;
+		  } else if (strcmp(name, "strings area") == 0) {
+			  debug_syms.stringarea = address;
 		  }
 	      }
 	    pos++;
@@ -1387,107 +1473,107 @@ int debug_eval_lex(void)
 {
   int start;
 
-  if (expr[expr_pos] == 0)
+  if (debug_expr[debug_expr_pos] == 0)
     return 0;
 
-  while (expr[expr_pos] == ' ')
-    expr_pos++;
+  while (debug_expr[debug_expr_pos] == ' ')
+    debug_expr_pos++;
 
-  start = expr_pos;
+  start = debug_expr_pos;
 
-  if ((expr[expr_pos] >= 'A' && expr[expr_pos] <= 'Z') ||
-      (expr[expr_pos] >= 'a' && expr[expr_pos] <= 'z'))
+  if ((debug_expr[debug_expr_pos] >= 'A' && debug_expr[debug_expr_pos] <= 'Z') ||
+      (debug_expr[debug_expr_pos] >= 'a' && debug_expr[debug_expr_pos] <= 'z'))
     {
       int x;
 
       /* IDENTIFIER */
-      while ((expr[expr_pos] >= 'A' && expr[expr_pos] <= 'Z') ||
-	     (expr[expr_pos] >= 'a' && expr[expr_pos] <= 'z') ||
-	     (expr[expr_pos] >= '0' && expr[expr_pos] <= '9') ||
-	     expr[expr_pos] == '_')
+      while ((debug_expr[debug_expr_pos] >= 'A' && debug_expr[debug_expr_pos] <= 'Z') ||
+	     (debug_expr[debug_expr_pos] >= 'a' && debug_expr[debug_expr_pos] <= 'z') ||
+	     (debug_expr[debug_expr_pos] >= '0' && debug_expr[debug_expr_pos] <= '9') ||
+	     debug_expr[debug_expr_pos] == '_')
 	{
-	  expr_pos++;
+	  debug_expr_pos++;
 	}
 
-      yylval.str = malloc(sizeof(char)*(expr_pos-start+1));
-      for (x=start; x<expr_pos; x++)
+      yylval.str = malloc(sizeof(char)*(debug_expr_pos-start+1));
+      for (x=start; x<debug_expr_pos; x++)
 	{
-	  yylval.str[x-start] = expr[x];
+	  yylval.str[x-start] = debug_expr[x];
 	}
-      yylval.str[expr_pos-start] = 0;
+      yylval.str[debug_expr_pos-start] = 0;
 
       return IDENTIFIER;
     }
   
-  if (expr[expr_pos] >= '0' && expr[expr_pos] <= '9')
+  if (debug_expr[debug_expr_pos] >= '0' && debug_expr[debug_expr_pos] <= '9')
     {
       /* NUMBER */
       yylval.number = 0;
 
-      while (expr[expr_pos] >= '0' && expr[expr_pos] <= '9')
+      while (debug_expr[debug_expr_pos] >= '0' && debug_expr[debug_expr_pos] <= '9')
 	{
 	  yylval.number *= 10;
-	  yylval.number += expr[expr_pos] - '0';
-	  expr_pos++;
+	  yylval.number += debug_expr[debug_expr_pos] - '0';
+	  debug_expr_pos++;
 	}
       return NUMBER;
     }
 
-  if (expr[expr_pos] == '$')
+  if (debug_expr[debug_expr_pos] == '$')
     {
       /* NUMBER */
       yylval.number = 0;
-      expr_pos++;
+      debug_expr_pos++;
 
-      while ((expr[expr_pos] >= '0' && expr[expr_pos] <= '9') ||
-	     (expr[expr_pos] >= 'A' && expr[expr_pos] <= 'F') ||
-	     (expr[expr_pos] >= 'a' && expr[expr_pos] <= 'f'))
+      while ((debug_expr[debug_expr_pos] >= '0' && debug_expr[debug_expr_pos] <= '9') ||
+	     (debug_expr[debug_expr_pos] >= 'A' && debug_expr[debug_expr_pos] <= 'F') ||
+	     (debug_expr[debug_expr_pos] >= 'a' && debug_expr[debug_expr_pos] <= 'f'))
 	{
 	  yylval.number *= 16;
-	  if (expr[expr_pos] >= '0' && expr[expr_pos] <= '9')
-	    yylval.number += expr[expr_pos] - '0';
-	  else if (expr[expr_pos] >= 'A' && expr[expr_pos] <= 'F')
-	    yylval.number += expr[expr_pos] - 'A' + 10;
-	  else if (expr[expr_pos] >= 'a' && expr[expr_pos] <= 'f')
-	    yylval.number += expr[expr_pos] - 'a' + 10;
+	  if (debug_expr[debug_expr_pos] >= '0' && debug_expr[debug_expr_pos] <= '9')
+	    yylval.number += debug_expr[debug_expr_pos] - '0';
+	  else if (debug_expr[debug_expr_pos] >= 'A' && debug_expr[debug_expr_pos] <= 'F')
+	    yylval.number += debug_expr[debug_expr_pos] - 'A' + 10;
+	  else if (debug_expr[debug_expr_pos] >= 'a' && debug_expr[debug_expr_pos] <= 'f')
+	    yylval.number += debug_expr[debug_expr_pos] - 'a' + 10;
 
-	  expr_pos++;
+	  debug_expr_pos++;
 	}
 
       return NUMBER;
     }
 
-  if (expr[expr_pos] == '-')
+  if (debug_expr[debug_expr_pos] == '-')
     {
-      if (expr[expr_pos+1] == '>')
+      if (debug_expr[debug_expr_pos+1] == '>')
 	{
-	  expr_pos+=2;
+	  debug_expr_pos+=2;
 	  return BYTEARRAY;
 	}
-      if (expr[expr_pos+1] == '-' && expr[expr_pos+2] == '>')
+      if (debug_expr[debug_expr_pos+1] == '-' && debug_expr[debug_expr_pos+2] == '>')
 	{
-	  expr_pos += 3;
+	  debug_expr_pos += 3;
 	  return WORDARRAY;
 	}
     }
 
-  if (expr[expr_pos] == '.')
+  if (debug_expr[debug_expr_pos] == '.')
     {
-      if (expr[expr_pos+1] == '&')
+      if (debug_expr[debug_expr_pos+1] == '&')
 	{
-	  expr_pos+=2;
+	  debug_expr_pos+=2;
 	  return PROPADDR;
 	}
-      if (expr[expr_pos+1] == '#')
+      if (debug_expr[debug_expr_pos+1] == '#')
 	{
-	  expr_pos+=2;
+	  debug_expr_pos+=2;
 	  return PROPLEN;
 	}
     }
 
-  expr_pos++;
-  if (expr[expr_pos-1] < 256)
-    return expr[expr_pos-1];
+  debug_expr_pos++;
+  if (debug_expr[debug_expr_pos-1] < 256)
+    return debug_expr[debug_expr_pos-1];
   return '?';
 }
 
@@ -1514,3 +1600,52 @@ char* debug_print_value(ZWord value, char* type)
   return res;
 }
 
+void debug_set_bp_handler(debug_breakpoint_handler handler) {
+	static int initialised = 0;
+	int x;
+	
+	bp_handler = handler;
+	
+	if (!initialised) {
+		initialised = 1;
+		
+		for (x=0; x<debug_syms.nroutines; x++) {
+			debug_set_breakpoint(debug_syms.routine[x].start+1,
+								 0, 1);
+		}
+	}
+}
+
+void debug_set_temp_breakpoints(debug_step_type step) {
+	int ln;
+	ZFrame* frame;
+	
+	if (addr.routine != NULL) {	
+		ln = addr.line_no;
+		if (ln != -1) {
+			ln++;
+		}
+	
+		if (ln >= addr.routine->nlines) {
+			ln = -1;
+		}
+		
+		if (step == debug_step_into) {
+			stepinto = 1;
+		}
+		
+		/* Set a breakpoint on each line... */
+		if (step != debug_step_out) {
+			for (ln = 0; ln < addr.routine->nlines; ln++) {
+				debug_set_breakpoint(addr.routine->line[ln].address, 1, 0);
+			}
+		}
+	}
+	
+	/* Set a breakpoint on the return location of this function */
+	if (machine.stack.current_frame) {
+		for (frame = machine.stack.current_frame->last_frame; frame != NULL; frame = frame->last_frame) {
+			frame->break_on_return = 1;
+		}
+	}
+}
